(*******************************************************************)
(*  This is part of SymbolicKAT, it is distributed under the       *)
(*  terms of the GNU Lesser General Public License version 3       *)
(*           (see file LICENSE for more details)                   *)
(*                                                                 *)
(*  Copyright 2014: Damien Pous. (CNRS, LIP - ENS Lyon, UMR 5668)  *)
(*******************************************************************)

open Common
open Hashcons

(** BDD nodes are hashconsed to ensure unicity, whence the following two-levels type *)
type 'a node = 'a descr hash_consed
and 'a descr = V of 'a | N of key * 'a node * 'a node

type 'a mem = 'a descr Hashcons.t

let init () = Hashcons.create 57

(** specific hashconsing function *)
let hashcons =
  let hash = function V x -> Hashtbl.hash x | N(a,l,r) -> Hashtbl.hash (a,l.hkey,r.hkey) in
  let equal x y = match x,y with
    | V x, V y -> x=y
    | N(a,l,r), N(a',l',r') -> a=a' && l==l' && r==r'
    | _ -> false
  in fun x -> hashcons hash equal x

let constant mem v = hashcons mem (V v)

let node mem b x y = if x==y then x else hashcons mem (N(b,x,y))

(* utilities for memoised recursive function over hashconsed values *)

let gc_caches = ref false

let memo_rec1 ?(gc=false) f =
  let t = ref Hmap.empty in
  if gc then ignore (Gc.create_alarm (fun () -> if !gc_caches then t := Hmap.empty));
  (* TODO: put the above alarm before the gc... *)
  let rec aux x =
    try Hmap.find x !t 
    with Not_found -> 
      let y = f aux x in 
      t := Hmap.add x y !t; y
  in aux

let memo_rec2 ?(gc=false) f =
  let t = ref Hmap.empty in
  if gc then ignore (Gc.create_alarm (fun () -> if !gc_caches then t := Hmap.empty));
  (* TODO: put the above alarm before the gc... *)
  let rec aux x y =
    try let tx = Hmap.find x !t in 
	try Hmap.find y !tx
	with Not_found ->
	  let z = f aux x y in
	  tx := Hmap.add y z !tx; z
    with Not_found -> 
      let z = f aux x y in 
      let tx = ref Hmap.empty in
      t := Hmap.add x tx !t; 
      tx := Hmap.add y z !tx;
      z
  in aux

let memo_rec2' f =
  let t = ref Hmap.empty in
  fun p g -> 
  let rec aux at x y =
    let aux b c x y = aux ((b,c)::at) x y in
    try let tx = Hmap.find x !t in 
	try Hmap.find y !tx
	with Not_found ->
	  let z = f aux at p g x y in
	  tx := Hmap.add y z !tx; z
    with Not_found -> 
      let z = f aux at p g x y in 
      let tx = ref Hmap.empty in
      t := Hmap.add x tx !t; 
      tx := Hmap.add y z !tx;
      z
  in aux []


let unary ?gc mem f =
  memo_rec1 ?gc (fun app x ->
    match x.node with
      | V v -> constant mem (f v)
      | N(b,l,r) -> node mem b (app l) (app r)
  )

let binary ?gc mem f =
  memo_rec2 ?gc (fun app x y -> 
    match x.node, y.node with
      | V v, V w -> constant mem (f v w)
      | V _, N(b,l,r) -> node mem b (app x l) (app x r)
      | N(b,l,r), V _ -> node mem b (app l y) (app r y)
      | N(b,l,r), N(b',l',r') -> 
	match compare b b' with
	  | 0 -> node mem b  (app l l') (app r r')
	  | 1 -> node mem b' (app x l') (app x r')
	  | _ -> node mem b  (app l y ) (app r y )
  )

let times mem z = 
  memo_rec2 (fun times f x ->
    match f.node, x.node with
      | V true, _ -> x
      | V false, _ -> z
      | N(b,l,r), V _ -> node mem b (times l x) (times r x)
      | N(b,l,r), N(b',l',r') ->
	match compare b b' with
	  | 0 -> node mem b  (times l l') (times r r')
	  | 1 -> node mem b' (times f l') (times f r')
	  | _ -> node mem b  (times l x ) (times r x )
  )

let pairs f =
  memo_rec2' (fun app at p g x y -> 
    Stats.incr "iter_pairs calls";
    match x.node, y.node with
      | V v, V w -> f ((p,at)::g, v,w)
      | V _, N(b,l,r) -> app b false x l; app b true x r
      | N(b,l,r), V _ -> app b false l y; app b true r y
      | N(b,l,r), N(c,l',r') -> 
	match compare b c with
	  | 0 -> app b false l l'; app b true r r'
	  | 1 -> app c false x l'; app c true x r'
	  | _ -> app b false l y ; app b true r y 
  )

let pairs' f =
  let m = ref Hmap.empty in
  let link x y = m := Hmap.add x y !m in
  let get x = try Some (Hmap.find x !m) with Not_found -> None in
  let rec repr x = 
    match get x with
      | None -> x
      | Some y -> match get y with
	  | None -> y
	  | Some z -> link x z; repr z
  in 
  fun p g -> 
  let rec unify at x y =
    let unify b c x y = unify ((b,c)::at) x y in
    let x = repr x in
    let y = repr y in
    if x!=y then (Stats.incr "unify calls";
    match x.node, y.node with
      | V a, V b -> link x y; f ((p,at)::g, a, b)
      | N(b,l,r), V _ -> link x y; unify b false l y; unify b true r y
      | V _, N(b,l,r) -> link y x; unify b false x l; unify b true x r
      | N(b,l,r), N(c,l',r') -> 
	match compare b c with
	  | 0 -> link x y; unify b false l l'; unify b true r r'
	  | 1 -> link y x; unify c false x l'; unify c true x r'
	  | _ -> link x y; unify b false l y ; unify b true r y )
  in unify []


type formula = bool node
let m = init ()
let bot = constant m false
let top = constant m true
let var b = node m b bot top
let rav b = node m b top bot

let neg = unary ~gc:true m not
let iff = binary ~gc:true m (=)
let xor = binary ~gc:true m (<>)

(* let dsj = binary m (||) *)
let dsj = memo_rec2 ~gc:true (fun dsj x y ->
  match x.node, y.node with
    | V true, _ | _, V false -> x
    | _, V true | V false, _ -> y
    | N(b,l,r), N(b',l',r') ->
      match compare b b' with
	| 0 -> node m b  (dsj l l') (dsj r r')
	| 1 -> node m b' (dsj x l') (dsj x r')
	| _ -> node m b  (dsj l y ) (dsj r y )
)

(* let cnj = binary m (&&) *)
let cnj = memo_rec2 ~gc:true (fun cnj x y ->
  match x.node, y.node with
    | V false, _ | _, V true -> x
    | _, V false | V true, _ -> y
    | N(b,l,r), N(b',l',r') ->
      match compare b b' with
	| 0 -> node m b  (cnj l l') (cnj r r')
	| 1 -> node m b' (cnj x l') (cnj x r')
	| _ -> node m b  (cnj l y ) (cnj r y )
)

let rec witness n =
  match n.node with
    | V _ -> []
    | N(a,l,r) -> if l==bot then (a,true)::witness r else (a,false)::witness l

let rec print_formula f x = match x.node with
  | V false -> Format.fprintf f "0"
  | V true -> Format.fprintf f "1"
  | N(a,l,r) when l==bot && r==top -> Format.fprintf f "%c" a
  | N(a,l,r) when l==top && r==bot -> Format.fprintf f "!%c" a
  | N(a,l,r) -> Format.fprintf f "(if %c %a %a)" a print_formula r print_formula l

let unary m f = unary ~gc:false m f
let binary m f = binary ~gc:false m f
