(*******************************************************************)
(*  This is part of SymbolicKAT, it is distributed under the       *)
(*  terms of the GNU Lesser General Public License version 3       *)
(*           (see file LICENSE for more details)                   *)
(*                                                                 *)
(*  Copyright 2014: Damien Pous. (CNRS, LIP - ENS Lyon, UMR 5668)  *)
(*******************************************************************)

(** Binary Decision Diagrams (BDDs) *)

open Common

(** Multi-Terminal BDD nodes with leaves in ['a] *)
type 'a node

(** memories for BDDs nodes with leaves in ['a] (required to ensure unicity) *)
type 'a mem

(** empty memory *)
val init: unit -> 'a mem

(** constant BDD (leaf) *)
val constant: 'a mem -> 'a -> 'a node

(** applying a unary function to a BDD *)
val unary: 'b mem -> ('a -> 'b) -> 'a node -> 'b node

(** applying a binary function to two BDDs (the standard "apply" function) *)
val binary: 'c mem -> ('a -> 'b -> 'c) -> 'a node -> 'b node -> 'c node

(** iterating over the pairs of values reachable from two given BDD nodes *)
val pairs: (gstring_ * 'a * 'b -> unit) -> var -> gstring_ -> 'a node -> 'b node -> unit

(** iterating over the pairs of values reachable from two given BDD
    nodes, up to equivalence.
    (This function is the key function for defining the symbolic version
    of Hopcroft and Karp algorithm; it uses a disjoint set forest
    internally.) *)
val pairs': (gstring_ * 'a * 'a -> unit) -> var -> gstring_ -> 'a node -> 'a node -> unit


(** symbolic Boolean formula *)
type formula = bool node

(** the various connectives  *)
val neg: formula -> formula
val dsj: formula -> formula -> formula
val cnj: formula -> formula -> formula
val xor: formula -> formula -> formula
val iff: formula -> formula -> formula

(** the two constants *)
val bot: formula
val top: formula

(** literals: a single variable, or the negation of a variable *)
val var: key -> formula
val rav: key -> formula

(** should we clean the Boolean BDDs caches *)
val gc_caches: bool ref

(** witness assignation for a satisfiable formula *)
val witness: formula -> (key*bool) list

(** [times m z f n] multiplies the BDD [n] by the formula [f], using [z] where [f] is false *)
val times: 'a mem -> 'a node -> formula -> 'a node -> 'a node

(** pretty-printing a symbolic formula (warning: exponential) *)
val print_formula: Format.formatter -> formula -> unit
