(*******************************************************************)
(*  This is part of SymbolicKAT, it is distributed under the       *)
(*  terms of the GNU Lesser General Public License version 3       *)
(*           (see file LICENSE for more details)                   *)
(*                                                                 *)
(*  Copyright 2014: Damien Pous. (CNRS, LIP - ENS Lyon, UMR 5668)  *)
(*******************************************************************)

type test =
  | Dsj of test * test 
  | Cnj of test * test 
  | Neg of test
  | Top
  | Bot
  | Prd of Common.key

let rec test_to_formula = function
  | Dsj(x,y) -> Bdd.dsj (test_to_formula x) (test_to_formula y)
  | Cnj(x,y) -> Bdd.cnj (test_to_formula x) (test_to_formula y)
  | Neg x -> Bdd.neg (test_to_formula x)
  | Top -> Bdd.top
  | Bot -> Bdd.bot
  | Prd i -> Bdd.var i

type 'b expr_ =
  | Pls of 'b expr_ * 'b expr_
  | Dot of 'b expr_ * 'b expr_
  | Str of 'b expr_
  | Tst of 'b
  | Var of Common.var

type expr = test expr_ 
type expr' = Bdd.formula expr_

let rec epsilon = function
  | Pls(x,y) -> Bdd.dsj (epsilon x) (epsilon y)
  | Dot(x,y) -> Bdd.cnj (epsilon x) (epsilon y)
  | Str x -> Bdd.top
  | Var i -> Bdd.bot
  | Tst t -> t

let zer = Tst Bdd.bot
let one = Tst Bdd.top
let str x = Str x
let tst t = Tst(t)
let var i = Var(i)

let rec pls e f = 
  match e,f with
    | Tst t, Tst t' -> Tst (Bdd.dsj t t')
    | Tst e, x | x, Tst e when e==Bdd.bot -> x
    | Tst a, Pls(Tst b,q) -> Pls(Tst (Bdd.dsj a b), q)
    | Pls(x,h),Pls(y,k) -> 
      if x<y then Pls(x,pls h f)
      else if x=y then Pls(x,pls h k)
      else Pls(y,pls e k)
    | x,Pls(y,q) | Pls(y,q),x -> 
      if x<y then Pls(x,Pls(y,q))
      else if x=y then Pls(y,q)
      else Pls(y,pls x q)
    | x,y -> 
      if x<y then Pls(x,y)
      else if x=y then x
      else Pls(y,x)

let rec dot e f =
  match e,f with
    | Tst e, x | x, Tst e when e==Bdd.top -> x
    | Tst e, _ | _, Tst e when e==Bdd.bot -> zer
    | Tst t, Tst t' -> Tst (Bdd.cnj t t')
    | Tst a, Dot(Tst b,q) -> 
      let c = Bdd.cnj a b in 
      if c==Bdd.bot then zer 
      else Dot(Tst c, q)
    | Dot(x,h),f -> Dot(x,dot h f)
    | _ -> Dot(e,f)

let rec expr' = function
  | Pls(x,y) -> pls (expr' x) (expr' y)
  | Dot(x,y) -> dot (expr' x) (expr' y) 
  | Str x -> str (expr' x)
  | Tst t -> Tst (test_to_formula t)
  | Var i -> Var i

let ssf_str =
  (* ensuring strict star form *)
  let plus_but_one a b = match a,b with
    | (Tst t, x | x, Tst t) when t==Bdd.top -> x
    | _ -> pls a b
  in
  let plus_but_tst a b = match a,b with
    | Tst t, x | x, Tst t -> x
    | _ -> pls a b
  in
  let rec remove e = 
    if epsilon e == Bdd.top then
      match e with
	| Pls(a,b) -> plus_but_one (remove a) (remove b)
	| Dot(a,b) -> plus_but_tst (remove a) (remove b)
	| Str e -> e
	| e -> e 
    else e
  in 
  fun x -> match remove x with
    | Tst t -> one
    | x -> Str(x)

let rec ssf = function
  | Pls(x,y) -> pls (ssf x) (ssf y)
  | Dot(x,y) -> dot (ssf x) (ssf y) 
  | Str x -> ssf_str (ssf x)
  | v -> v

let rec print_test0 f = function
  | Dsj(x,y) -> Format.fprintf f "%a+%a" print_test0 x print_test0 y
  | x -> print_test1 f x
and print_test1 f = function
  | Cnj(x,y) -> Format.fprintf f "%a%a" print_test2 x print_test1 y
  | x -> print_test2 f x
and print_test2 f = function
  | (Dsj(_,_) | Cnj(_,_)) as x -> Format.fprintf f "(%a)" print_test0 x
  | Neg x -> Format.fprintf f "!%a" print_test2 x
  | Top -> Format.fprintf f "1"
  | Bot -> Format.fprintf f "0"
  | Prd i -> Format.fprintf f "%c" i

let rec print_expr0 f = function
  | Pls(x,y) -> Format.fprintf f "%a+%a" print_expr0 x print_expr0 y
  | x -> print_expr1 f x
and print_expr1 f = function
  | Dot(x,y) -> Format.fprintf f "%a%a" print_expr2 x print_expr1 y
  | x -> print_expr2 f x
and print_expr2 f = function
  | (Pls(_,_)|Dot(_,_)) as x -> Format.fprintf f "(%a)" print_expr0 x
  | Str x -> Format.fprintf f "%a*" print_expr2 x
  | Var i -> Format.fprintf f "%c" i
  | Tst t -> print_test0 f t

let rec print_expr'0 f = function
  | Pls(x,y) -> Format.fprintf f "%a+%a" print_expr'0 x print_expr'0 y
  | x -> print_expr'1 f x
and print_expr'1 f = function
  | Dot(x,y) -> Format.fprintf f "%a%a" print_expr'2 x print_expr'1 y
  | x -> print_expr'2 f x
and print_expr'2 f = function
  | (Pls(_,_)|Dot(_,_)) as x -> Format.fprintf f "(%a)" print_expr'0 x
  | Str x -> Format.fprintf f "%a*" print_expr'2 x
  | Var i -> Format.fprintf f "%c" i
  | Tst t -> Bdd.print_formula f t

let print_test = print_test0
let print_expr = print_expr0
let print_expr' = print_expr'0

let random_expr k v n =
  let k = Array.of_list (Top::List.map (fun x -> Prd x) k@List.map (fun x -> Neg (Prd x)) k) in
  let nk = Array.length k in
  let v = Array.of_list v in
  let nv = Array.length v in
  let rec expr = function
  | 0 -> if Random.int (nk+nv) < nk then Tst k.(Random.int nk) else Var v.(Random.int nv)
  | n -> 
    if Random.int 4 = 0 then Str (expr (n-1))
    else let i = Random.int n in 
	 let l,r = expr i, expr (n-i-1) in
	 if Random.bool() then Pls(l,r) else Dot(l,r)
  in fun () -> expr n

let random_full_expr k v n =
  let full = Str(List.fold_left (fun z i -> Pls (z,Var i)) (Var (List.hd v)) (List.tl v)) in
  let expr = random_expr k v n in
  fun () -> Pls(full,expr ())
