(*******************************************************************)
(*  This is part of SymbolicKAT, it is distributed under the       *)
(*  terms of the GNU Lesser General Public License version 3       *)
(*           (see file LICENSE for more details)                   *)
(*                                                                 *)
(*  Copyright 2014: Damien Pous. (CNRS, LIP - ENS Lyon, UMR 5668)  *)
(*******************************************************************)

open Common

(* symbolic without upto techniqes*)
module Symb = Equiv.Symbolic(Queues.BFS)(Upto.SET) 

(* symbolic hopcroft and Karp (with dsf) *)
module SymbHK = Equiv.SymbolicHK(Queues.BFS)

(* quiet or not, stats or not *)
let quiet = ref !Sys.interactive
let stats = ref false

(* do we use Hopcroft and Karp's version? *)
let hk = ref false
let equiv_ a x y = 
  match if !hk then SymbHK.equiv a x y else Symb.equiv a x y with
    | None -> None
    | Some (a,b,w) -> Some (List.rev (Bdd.witness (Bdd.xor a b)), w)

(* do we put expressions in strict star form first? *)
let ssf = ref false
let expr' x = 
  let x = Kat.expr' x in
  if !ssf then Kat.ssf x else x

(* which construction to use *)
let construction = ref `Antimirov
let set x () = construction := x

(* Antimirov' construction *)
let antimirov x y =
  let a,f = Determinisation.reindexed_dfa (Antimirov.nfa()) in
  equiv_ a (f (Set.single x)) (f (Set.single y))

(* Brzozowski's construction *)
let brzozowski x y = 
  equiv_ (Brzozowski.dfa()) x y

(* Ilie and Yu's construction *)
let ilie_yu x y =
  let a,x,y = IlieYu.enfa x y in
  let a = Epsilon.remove a in
  let a,f = Determinisation.reindexed_dfa a in
  equiv_ a (f (Set.single x)) (f (Set.single y))

(* equivance check *)
let equiv x y = 
  let x,y = expr' x, expr' y in
  match !construction with
    | `Antimirov -> antimirov x y
    | `Brzozowski -> brzozowski x y
    | `IlieYu -> ilie_yu x y

(* equivalence/inclusion/difference check *)
let check m x y =
  let x = Parse.expr x in
  let y = Parse.expr y in
  let z = Kat.Pls(x,y) in
  match m with
    | `E -> assert(None = equiv x y)
    | `L -> assert(None <> equiv x y); assert(None = equiv z y)
    | `R -> assert(None <> equiv x y); assert(None = equiv z x)
    | `D -> assert(None <> equiv x y); assert(None <> equiv z y); assert(None <> equiv z x)

(* entry point, for standalone program *)
let main() =
  let args = Arg.align [
    "-q", Arg.Set quiet, " quiet mode";
    "-s", Arg.Set stats, " print statistics";
    "-gc", Arg.Set Bdd.gc_caches, " clear the Boolean BDD caches from time to time";
    "-ssf", Arg.Set ssf, " put expressions in strict star form first";
    "-hk", Arg.Set hk, " use symbolic version of Hopcroft and Karp's algorithm";
    "-ant", Arg.Unit (set `Antimirov), " use Antimirov' partial derivatives [default]";
    "-brz", Arg.Unit (set `Brzozowski), " use Brzozowski's derivatives";
    "-iyu", Arg.Unit (set `IlieYu), " use Ilie and Yu's construction"    
  ] in
  let doc = Format.sprintf "usage: %s [options] [x y]

This programs checks equivalence of KAT expressions [x] and [y].
See http://perso.ens-lyon.fr/damien.pous/symbolickat/ for a complete description.

The syntax for KAT expressions is the following:
 - atomic tests are characters from `a' to `j'
 - atomic Kleene elements are characters from `k' to `z'
 - multiplication or Boolean conjunction is implicit, by juxtaposition
 - addition or Boolean disjunction is `+'
 - Kleene star is postfix `*'
 - Boolean negation is prefix `!'
 - zero and one are `0' and `1'
Typically, `(ap+q)*!a' is a wellformed expression.

Options are the following ones:" Sys.argv.(0) in
  let usage() = Arg.usage args doc in
  let anon =
    let one = ref None in
    let two = ref false in
    fun y -> 
      match !one with None -> one := Some y | Some x ->
	if !two then (usage(); exit 2);
	two := true;
	let e = match equiv (Parse.expr x) (Parse.expr y) with
	  | None -> if not !quiet then Format.printf "true\n"; 0
	  | Some g -> if not !quiet then Format.printf "false: %a\n" print_gstring g; 1
	in
	if !stats then Stats.print Format.std_formatter;
	exit e
  in Arg.parse args anon doc

let _ = if not !Sys.interactive then main()
